import sys
from sly import Lexer

class HorstLexer(Lexer):
    tokens = { ADD, AS, COMMENT, CLONE, CREATE, DROP, FIELDS, FIELD_TYPE, ID, ON
        , SHOW_ALL, TABLE, TABLES, TYPE
        , WITH, WITHOUT }
    ignore = ' \t'
    literals = { ',', }

    # Identifiers
    ID = r'[a-z_][a-z0-9_.]*'
    COMMENT = r'--.*'
    
    # keywords aka terminal identifiers
    ID['add'] = ADD
    ID['as'] = AS
    ID['create'] = CREATE
    ID['clone'] = CLONE
    ID['drop'] = DROP
    ID['fields'] = FIELDS
    ID['integer'] = FIELD_TYPE
    ID['string'] = FIELD_TYPE
    ID['timestamp'] = FIELD_TYPE
    ID['on'] = ON
    ID['show_all'] = SHOW_ALL
    ID['table'] = TABLE
    ID['tables'] = TABLES
    ID['type'] = TYPE
    ID['with'] = WITH
    ID['without'] = WITHOUT

    # Line number tracking
    @_(r'\n+')
    def ignore_newline(self, t):
        self.lineno += t.value.count('\n')

    def error(self, token):
        sys.exit(f"ERROR during lexical analysis: Bad character '{token.value[0]}'")


import copy
import sys
from sly import Parser

from horst_l import HorstLexer
from horst_c import HorstTable
# from horst.horst_l import HorstLexer
# from horst.horst_c import HorstTable

class HorstParser(Parser):
    tokens = HorstLexer.tokens
    tables = {}
    wrap_keywords = HorstTable('tool').wrap_keywords

    def error(self, token):
        print(f".......... ERROR - during parsing: Type '{token.type}' with value '{token.value}'")

# ++++++++++++++++++++++++++++++++++++++++++
# +++ statement
# ++++++++++++++++++++++++++++++++++++++++++

    @_('table_declaration')
    def statement(self, p) -> str:
        return p.table_declaration

    @_('clone_table')
    def statement(self, p) -> str:
        return p.clone_table

    @_('DROP TABLES')
    def statement(self, p):
        sql_code = ''
        first_table = True
        for table_name in self.tables:
            if first_table:
                first_table = False
            else:
                sql_code = sql_code + '\n'
            sql_code = sql_code + self.tables[table_name].drop_table()
        return sql_code

    @_('list_fields')
    def statement(self, p) -> str:
        return p.list_fields

    @_('join_fields')
    def statement(self, p) -> str:
        return p.join_fields

    @_('COMMENT')
    def statement(self, p) -> str:
        return ''

# ++++++++++++++++++++++++++++++++++++++++++
# +++ table_declaration
# ++++++++++++++++++++++++++++++++++++++++++

    @_('error CREATE ')
    def table_declaration(self, p):
        sys.exit(f".......... ERROR - a table declaration begins with 'CREATE' ")

    @_('CREATE TABLE ID table_instance field_list')
    def table_declaration(self, p) -> str:
        return self.tables[self.table_name].create_table()

    @_('')
    def table_instance(self, p):
        self.table_name = p[-1]
        self.tables[self.table_name] = HorstTable(self.table_name)

    @_('SHOW_ALL')
    def table_declaration(self, p):
        for table_instance in self.tables.values():
            table_instance.show()

# ++++++++++++++++++++++++++++++++++++++++++
# +++ clone_table
# ++++++++++++++++++++++++++++++++++++++++++

    @_('clone_this_table add_field_list')
    def clone_table(self, p):
        return self.tables[self.table_name].create_table()

    @_('clone_this_table WITHOUT field_names')
    def clone_table(self, p):
        without_list = []
        for field in self.tables[self.table_name].field_list:
            if field["name"] not in p.field_names:
                without_list.append(field)
        self.tables[self.table_name].field_list = without_list
        return self.tables[self.table_name].create_table()

    @_('')
    def add_field_list(self, p):
        pass

    @_('ADD field_list')
    def add_field_list(self, p):
        pass

    @_('CLONE TABLE ID AS ID table_id')
    def clone_this_table(self, p):
        self.tables[self.table_name] = copy.deepcopy(self.tables[p.ID0])
        self.tables[self.table_name].table_name = self.table_name

    @_('')
    def table_id(self, p):
        self.table_name = p[-1]

# ++++++++++++++++++++++++++++++++++++++++++
# +++ list_fields
# ++++++++++++++++++++++++++++++++++++++++++

    @_('table_fields fields_with')
    def list_fields(self, p) -> str:
        return self.tables[p.table_fields['name']].fields(table_alias=p.table_fields['alias']
        , fields_out=p.fields_with['fields_out'], show_type=p.fields_with['show_type'])

    @_('table_fields')
    def list_fields(self, p) -> str:
        return self.tables[p.table_fields['name']].fields(table_alias=p.table_fields['alias'])

    @_('WITH TYPE')
    def fields_with(self, p) -> str:
        return {'show_type':True, 'fields_out':[]}
        # return self.tables[p.ID].fields(show_type=True)

    @_('WITH TYPE WITHOUT field_names')
    def fields_with(self, p) -> str:
        return {'show_type':True, 'fields_out':p.field_names}

    @_('WITHOUT field_names')
    def fields_with(self, p) -> str:
        return {'show_type':False, 'fields_out':p.field_names}

    @_('ID FIELDS')
    def table_fields(self, p) -> str:
        return {'name':p.ID, 'alias':''}

    @_('ID table_alias FIELDS')
    def table_fields(self, p) -> str:
        return {'name':p.ID, 'alias':p.table_alias}

# ++++++++++++++++++++++++++++++++++++++++++
# +++ join_fields
# +++ join table_name {as table_alias} on table_name {as table_alias} {(with | without) field_names}
# ++++++++++++++++++++++++++++++++++++++++++

    @_('ID table_alias ON ID')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias if len(p.table_alias) >0 else p.ID0
        join_on_fields = ''
        first_field = True
        for field in self.tables[p.ID0].field_list:
            if not first_field:
                join_on_fields = join_on_fields + ' AND ' 
            join_on_fields = join_on_fields + p.ID1 + '.' + self.wrap_keywords(field["name"]) + ' = ' + tab1 \
            + '.' + self.wrap_keywords(field["name"]) 
            first_field = False
        return join_on_fields

    @_('ID table_alias ON ID table_alias')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias0 if len(p.table_alias0) >0 else p.ID0
        tab2 = p.table_alias1 if len(p.table_alias1) >0 else p.ID1
        join_on_fields = ''
        first_field = True
        for field in self.tables[p.ID0].field_list:
            if not first_field:
                join_on_fields = join_on_fields + ' AND ' 
            join_on_fields = join_on_fields + tab2 + '.' + self.wrap_keywords(field["name"]) + ' = ' + tab1 \
            + '.' + self.wrap_keywords(field["name"]) 
            first_field = False
        return join_on_fields

    @_('ID table_alias ON ID WITH field_names')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias if len(p.table_alias) >0 else p.ID0
        join_on_fields = ''
        first_field = True
        for field in p.field_names:
            if not first_field:
                join_on_fields = join_on_fields + ' AND ' 
            join_on_fields = join_on_fields + p.ID1 + '.' + self.wrap_keywords(field) + ' = ' \
            + tab1 + '.' + self.wrap_keywords(field)
            first_field = False
        return join_on_fields

    @_('ID table_alias ON ID table_alias WITH field_names')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias0 if len(p.table_alias0) >0 else p.ID0
        tab2 = p.table_alias1 if len(p.table_alias1) >0 else p.ID1
        join_on_fields = ''
        first_field = True
        for field in p.field_names:
            if not first_field:
                join_on_fields = join_on_fields + ' AND ' 
            join_on_fields = join_on_fields + tab2 + '.' + self.wrap_keywords(field) + ' = ' \
            + tab1 + '.' + self.wrap_keywords(field)
            first_field = False
        return join_on_fields

    @_('ID table_alias ON ID WITHOUT field_names')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias if len(p.table_alias) >0 else p.ID0
        join_on_fields = ''
        first_field = True
        for field in self.tables[p.ID0].field_list:
            if field["name"] not in p.field_names:
                if not first_field:
                    join_on_fields = join_on_fields + ' AND ' 
                join_on_fields = join_on_fields + p.ID1 + '.' + self.wrap_keywords(field["name"]) + ' = ' \
                + tab1 + '.' + self.wrap_keywords(field["name"]) 
                first_field = False
        return join_on_fields

    @_('ID table_alias ON ID table_alias WITHOUT field_names')
    def join_fields(self, p) -> str:
        tab1 = p.table_alias0 if len(p.table_alias0) >0 else p.ID0
        tab2 = p.table_alias1 if len(p.table_alias1) >0 else p.ID1
        join_on_fields = ''
        first_field = True
        for field in self.tables[p.ID0].field_list:
            if field["name"] not in p.field_names:
                if not first_field:
                    join_on_fields = join_on_fields + ' AND ' 
                join_on_fields = join_on_fields + tab2 + '.' + self.wrap_keywords(field["name"]) + ' = ' \
                + tab1 + '.' + self.wrap_keywords(field["name"]) 
                first_field = False
        return join_on_fields

# ++++++++++++++++++++++++++++++++++++++++++
# +++ common methods
# ++++++++++++++++++++++++++++++++++++++++++

    @_('ID')
    def field_names(self, p) -> str:
        element = [p.ID,]
        return element

    @_('ID "," field_names')
    def field_names(self, p) -> str:
        element = [p.ID,]
        field_names = p.field_names
        field_names.append(element[0])
        return field_names

    @_('AS ID')
    def table_alias(self, p) -> str:
        return f"{p.ID}"

    @_('field_definition')
    def field_list(self, p):
        pass

    @_('field_definition "," field_list')
    def field_list(self, p):
        pass

    @_('ID FIELD_TYPE')
    def field_definition(self, p):
        self.tables[self.table_name].add_field(name=p.ID, type=p.FIELD_TYPE)



class HorstTable():

    sql_critical_keywords = {'group'}

    def __init__(self, table_name):
        self.table_name = table_name
        self.field_list = []
        self.temp_table = 'TEMP'

    def add_field(self, name, type):
        field = {}
        field["name"] = name
        field["type"] = 'INT64' if type == 'integer' else type
        self.field_list.append(field)

    def create_table(self) -> str:
        table_list = ''
        for field in self.field_list:
            table_list = table_list + self.wrap_keywords(field["name"]) +' '+ field["type"] + ', '
        # uups, one comma 2 much
        table_list = table_list[:-2]
        sql_create_string = f"CREATE {self.temp_table} TABLE {self.table_name} ({table_list});"
        return sql_create_string        

    def drop_table(self) -> str:
        sql_code = f"DROP TABLE {self.table_name};"
        return sql_code

    def fields(self, fields_out=[], show_type=False, table_alias='') -> str:
        first_field = True
        for field in self.field_list:
            if field["name"] in fields_out:
                pass
            else:
                if first_field:
                    select_fields = (table_alias+'.' if table_alias != '' else '') + self.wrap_keywords(field["name"]) \
                    + ((' ' + field["type"]) if show_type else '')
                    first_field = False
                else:
                    select_fields = select_fields + ', ' + (table_alias+'.' if table_alias != '' else '') \
                    + self.wrap_keywords(field["name"]) + ((' ' + field["type"]) if show_type else '')
        return select_fields

    def show(self):
        print("..............................")
        print(f"table {self.table_name} has some fields {self.field_list} ")

    def wrap_keywords(self, key) -> str:
        if key in self.sql_critical_keywords:
            return "`" + key + "`"
        else:
            return key




""" 
    Extended BNF of horst

    statement = table_declaration 
        | clone_table
        | drop tables 
        | list_fields
        | join_fields
        | comment

    table_declaration = create table table_name field_list
    clone_table = clone table ID as ID {add field_list | without field_names}
    list_fields = table_name {as table_alias} fields {with type} {without field_names}
    join_fields = table_name {as table_alias} on table_name {as table_alias} {(with | without) field_names}
    comment = -- * 

    field_list = field_definition {, field_definition}
    field_definition = field_name field_type
    field_type = INTEGER | STRING | TIMESTAMP
    field_names = field_name {, field_name} 
    field_name = ID
    table_name = ID
    table_alias = ID
    ID = RegEx([a-zA-Z_][a-zA-Z0-9_]*) 

    code example 
    $ python horst.py this_horst_script
    << horst.this_horst_script.sql
    >> this_horst_script.sql

"""
import argparse
from sly import Parser

from horst_l import HorstLexer
from horst_p import HorstParser
# from horst.horst_l import HorstLexer
# from horst.horst_p import HorstParser

def new_line_on(string, length=90):

    def nl_position(string2search):
        position = string2search[:length].rfind(',')
        position_and = string2search[:length].lower().rfind('and') 
        position = position if position > position_and else position_and
        return position

    if len(string) > length:
        comma_position = nl_position(string)
        line = string[:comma_position]
        string = string[comma_position:]
        while len(string) > length:
            comma_position = nl_position(string)
            line = line + '\n' + string[:comma_position]
            string = string[comma_position:]
        if len(string) > 0:
            line = line + '\n' + string
    else:
        line = string
    return line

def horst():

    command_line_parser = argparse.ArgumentParser(description='Horst supports SQL.')
    command_line_parser.add_argument("file", help='SQL-file with horst code to process.')
    command_line_args = command_line_parser.parse_args()

    lexer = HorstLexer()
    parser = HorstParser()
    
    horst_path_file = command_line_args.file
    horst_path =  horst_path_file[:horst_path_file[::-1].find("/")+1]
    horst_file = horst_path_file[horst_path_file[::-1].find("/")+1:]

    EOF = ''
    read_horst = False

    with open(horst_path + 'horst.' + horst_file + '.sql', 'r') as hsql_file \
    , open(horst_path_file + '.sql', 'w') as sql_file:
        hsql_file_line = hsql_file.readline()

        while hsql_file_line != EOF:

            horst_start = hsql_file_line.find('{{')
            horst_end = hsql_file_line.find('}}')

            # start symbol detected
            if horst_start >= 0:
                read_horst = True
                horst_code = ''

            if read_horst:
                horst_code = horst_code + hsql_file_line[
                    horst_start +2 if horst_start >= 0 else None 
                    : horst_end  if horst_end >= 0 else None
                    ]

            # end symbol detected
            if horst_end >= 0:
                read_horst = False
                sql_code = parser.parse(lexer.tokenize(horst_code))
                sql_code = new_line_on(sql_code)
                hsql_file_line = hsql_file_line[:horst_end +2] +sql_code+ hsql_file_line[horst_end +2:]

            if horst_start > 0:
                sql_file_line = hsql_file_line[:horst_start if horst_start >= 0 else None]
                if horst_end >= 0:
                    sql_file_line = sql_file_line + hsql_file_line[horst_end +2 if horst_end >= 0 else None:]
                else:
                    sql_file_line = sql_file_line + '\n\r'
            else:
                sql_file_line = hsql_file_line[horst_end +2 if horst_end >= 0 else None:]

            if (horst_start > 0 or horst_end >= 0) and len(sql_file_line.strip(' \n\r')) > 0:
                sql_file.write(sql_file_line)
            elif not read_horst and horst_end < 0:
                sql_file.write(sql_file_line)

            hsql_file_line = hsql_file.readline()

if __name__ == '__main__':
    horst()


import setuptools
"""
    usage to build pypi package
    
    create distribution
    $ python setup.py sdist

    build wheel
    $ python setup.py bdist_wheel
"""
 
with open("README.md", "r") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name='horst',  
    version='0.3.14',
    author="Frank T. Maas",
    author_email="frank.maas@funkemedien.de",
    description="Do not repeat yourself in SQL scripts by listing the same fields again and again.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/FMG_ULM/horst",
    packages=["horst",],
    entry_points = {
        "console_scripts": ['aogl = aogl.aogl:main']
    },
    install_requires=[
        "sly"
        ,
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)